#include <iostream>
#include <string>
#include <TFile.h>
#include <TH1D.h>
#include <RooGlobalFunc.h>
#include <RooPlot.h>
#include <RooDataHist.h>
#include <RooRealVar.h>
#include <TCanvas.h>
#include "roothreshold.h"
#include <RooGaussian.h>
#include <RooFitResult.h>
#include <RooCategory.h>
#include <RooLandau.h>
#include <RooAddPdf.h>
#include <RooAbsReal.h>
#include <TRandom3.h>
#include <vector>
#include <sstream>
#include <TTree.h>

using namespace std;
using namespace RooFit;

typedef struct {
  double delta;
  double obs;
  double obs_err;
  double exp;
  double exp_err;
  double pull;
  int side_fit_status;
  int side_cov_status;
  int signal_fit_status;
  int signal_cov_status;
  double norm;
  double norm_err;
  double offset;
  double offset_err;
  double power;
  double power_err;
  double coeff1;
  double coeff1_err;
  double coeff2;
  double coeff2_err;
  double mean;
  double mean_err;
  double width;
  double width_err;
  TH1 *signal_histo;
  TH1 *sideband_histo;
  TCanvas *canvas;
} Result;

Result doFit(TH1*, TH1*, bool);
TH1* genMC(TH1*);
void printResult(Result);
void histoDiagnostic(TH1*, TH1*, TH1*, TH1*, string);
void storeData(vector<Result>&, string, TH1*, TH1*);

TH1* genMC(TRandom3 *rng, TH1 *input) {
  if(input == 0x0) return 0x0;

  // Clone the histogram
  TH1 *output = (TH1*) input->Clone();

  long nentries = output->GetEntries();

  // Empty it out
  output->Reset();

  for(long i = 0; i < nentries; i++) {
    output->Fill(rng->Gaus(100, 20));
  }

  return(output);

  // Get the x-axis to work with
  TAxis *xaxis = output->GetXaxis();

  // Now lets go bin by bin and populate it
  for(int i = 1; i < output->GetNbinsX(); i++) {
    // Get the i'th bin center
    double x = input->GetBinCenter(i);

    // Get the mean for the i'th bin from the original
    double input_count = input->GetBinContent(i);

    // Draw a number from the poisson distribution using this as the mean
    double output_count = rng->Poisson(input_count);

    // Set the content of the bin (keeping event counts)
    output->Fill(x, output_count);

    // Set the bin error
    output->SetBinError(i, sqrt(output_count));
  }

  output->SetEntries(input->GetEntries());

  return(output);
}

void storeData(vector<Result> &results, string filename, TH1 *sideband = 0x0, TH1 *signal = 0x0) {
  // Ok, so lets open a root file and create some leafs

  Result tmpres;

  TFile output(filename.c_str(), "RECREATE");
  TTree *tree = new TTree("results", "results", 0);

  tree->Branch("delta", &tmpres.delta, "delta/D");
  tree->Branch("obs", &tmpres.obs, "obs/D");
  tree->Branch("obs_err", &tmpres.obs_err, "obs_err/D");
  tree->Branch("exp", &tmpres.exp, "exp/D");
  tree->Branch("exp_err", &tmpres.exp_err, "exp_err/D");
  tree->Branch("pull", &tmpres.pull, "pull/D");
  tree->Branch("side_fit_status", &tmpres.side_fit_status, "side_fit_status/I");
  tree->Branch("side_cov_status", &tmpres.side_cov_status, "side_cov_status/I");
  tree->Branch("signal_fit_status", &tmpres.signal_fit_status, "signal_fit_status/I");
  tree->Branch("signal_cov_status", &tmpres.signal_cov_status, "signal_cov_status/I");
  tree->Branch("norm", &tmpres.norm, "norm/D");
  tree->Branch("norm_err", &tmpres.norm_err, "norm_err/D");
  tree->Branch("offset", &tmpres.offset, "offset/D");
  tree->Branch("offset_err", &tmpres.offset_err, "offset_err/D");
  tree->Branch("power", &tmpres.power, "power/D");
  tree->Branch("power_err", &tmpres.power_err, "power_err/D");
  tree->Branch("coeff1", &tmpres.coeff1, "coeff1/D");
  tree->Branch("coeff1_err", &tmpres.coeff1_err, "coeff1_err/D");
  tree->Branch("coeff2", &tmpres.coeff2, "coeff2/D");
  tree->Branch("coeff2_err", &tmpres.coeff2_err, "coeff2_err/D");
  tree->Branch("mean", &tmpres.mean, "mean/D");
  tree->Branch("mean_err", &tmpres.mean_err, "mean_err/D");
  tree->Branch("width", &tmpres.width, "width/D");
  tree->Branch("width_err", &tmpres.width_err, "width_err/D");

  TCanvas *tmpcan = 0x0;
  TH1     *tmpsignal = 0x0;
  TH1     *tmpsideband = 0x0;


  tree->Branch("canvas", "TCanvas", &tmpcan);

  tree->Branch("signal_histogram", &tmpsignal);
  tree->Branch("sideband_histogram", &tmpsideband);

  for(int i = 0; i < results.size(); i++) {
    tmpres      = results.at(i);
    tmpcan      = tmpres.canvas;
    tmpsignal   = tmpres.signal_histo;
    tmpsideband = tmpres.sideband_histo;

    stringstream name;
    name << "Fits/fit_" << i << ".eps";
    tmpcan->SaveAs(name.str().c_str());

    tree->Fill();
  }

  if(sideband != 0x0) {
    sideband->Write();
  }

  if(signal != 0x0) {
    signal->Write();
  }

  output.Write();
  output.Close();
}

int main() {
  // Place to store the outputs
  vector<Result> results;

  // Need to load the sideband distribution

  TFile input_sideband("sideband.root");
  TH1 *input_sideband_histo = (TH1*) input_sideband.FindObjectAny("data_0tag_rev1_bbcutinvlinear_jj")->Clone();

  // Need to load the signal region

  TFile input_signal("signal.root");
  TH1 *input_signal_histo = (TH1*) input_signal.FindObjectAny("data_0tag_rev1_bbcutlinear_jj")->Clone();

  TRandom3 *r = new TRandom3(0);

  // Lets just create some random distributions

  cout << "Generating ramdom Data Histograms from Poisson distribuions" << endl;

  long num_sideband = input_sideband_histo->GetEntries();
  long num_signal   = input_signal_histo->GetEntries();

  num_sideband = 1000;
  num_signal   = 500;

  for(int i = 1; i < input_sideband_histo->GetNbinsX(); i++) {
    input_sideband_histo->SetBinContent(i, 0);
    input_sideband_histo->SetBinError(i, 0);
  }

  for(int i = 1; i < input_signal_histo->GetNbinsX(); i++) {
    input_signal_histo->SetBinContent(i, 0);
    input_signal_histo->SetBinError(i, 0);
  }

  input_sideband_histo->Reset();
  input_sideband_histo->Sumw2();
  for(long i = 0; i < num_sideband; i++) {
    input_sideband_histo->Fill(r->Gaus(100, 20));
  }

  input_signal_histo->Reset();
  input_signal_histo->Sumw2();
  for(long i = 0; i < num_signal; i++) {
    input_signal_histo->Fill(r->Gaus(100, 20));
  }

//  delete input_signal_histo;
//  input_signal_histo = (TH1*) input_sideband_histo->Clone("blah");
//  input_signal_histo->Scale(0.5);

  cout << "Data Histogram generation compleated" << endl;

  long total = 1000;
  long maxrun = 100000;
  long i = 0;

  while(results.size() < total && i < maxrun) {
    i++;

    if(results.size()%100 == 0) {
      cerr << "Performing fit " << i << " of " << total << endl;
    }

    TH1 *tmp_side   = genMC(r, input_sideband_histo);
    TH1 *tmp_signal = genMC(r, input_signal_histo);

    stringstream tmpname;
    tmpname << i;

    if(i == 1) {
      histoDiagnostic(input_sideband_histo, input_signal_histo, tmp_side, tmp_signal, tmpname.str()); 
    }

    bool saveit = false;
    if(i == 1) {
      saveit = true;
    }

    Result tmpres = doFit(tmp_side, tmp_signal, saveit);

    results.push_back(tmpres);

//    delete tmp_side;
//    delete tmp_signal;
  }

  // Now print out the results information
  for(int i = 0; i < results.size(); i++) {
    Result tmpres = results.at(i);

    printResult(tmpres);
  }

  storeData(results, "results.root", input_sideband_histo, input_signal_histo);

  // Cleanup the memory used for the histograms
  for(int i = 0; i < results.size(); i++) {
    Result tmpres = results.at(i);

    delete tmpres.sideband_histo;
    delete tmpres.signal_histo;
    delete tmpres.canvas;

    tmpres.sideband_histo = 0x0;
    tmpres.signal_histo   = 0x0;
    tmpres.canvas         = 0x0;
  }
}

void histoDiagnostic(TH1 *input_sideband_histo, TH1 *input_signal_histo, TH1 *tmp_side, TH1 *tmp_signal, string ident) {
  // Generate three different plots
  //  1: input side vs tmp side
  //  2: input signal vs tmp signal
  //  3: difference between the two sideband
  //  4: diffeence between the two signal

  // Set the colors of the histograms
  tmp_side->SetLineColor(2);
  tmp_signal->SetLineColor(2);

  // Setup the canvas that I will be using
  TCanvas *can = new TCanvas("mycanvas");

  stringstream name;

  // Do the sidebands
  name.str("");
  name << "sideband_comparison_" << ident << ".eps";
  input_sideband_histo->Draw();
  tmp_side->Draw("SAME");
  can->SaveAs(name.str().c_str());

  // Do the signals
  name.str("");
  name << "signal_comparison_" << ident << ".eps";
  input_signal_histo->Draw();
  tmp_signal->Draw("SAME");
  can->SaveAs(name.str().c_str());

  // Do the sideband division
  name.str("");
  name << "sideband_division_" << ident << ".eps";
  TH1 *side_copy = (TH1*) input_sideband_histo->Clone();
  side_copy->Divide(tmp_side);
  side_copy->Draw();
  can->SaveAs(name.str().c_str());

  // Do the signal division
  name.str("");
  name << "signal_division_" << ident << ".eps";
  TH1 *signal_copy = (TH1*) input_signal_histo->Clone();
  signal_copy->Divide(tmp_signal);
  signal_copy->Draw();
  can->SaveAs(name.str().c_str());

  // Cleanup
  delete side_copy;
  delete signal_copy;
  delete can;
}

void printResult(Result tmp) {
  cout << "Fit Result" << endl;
  cout << "----------------------------------" << endl;
  cout << "  Backgrond Obs: " << tmp.obs << " +- " << tmp.obs_err << endl;
  cout << "  Backgrond Exp: " << tmp.exp << " +- " << tmp.exp_err << endl;
  cout << "  Delta        : " << tmp.delta << endl;
  cout << "  Pull         : " << tmp.pull << endl;
  cout << "  Fit Status   : " << tmp.side_fit_status << " (Sideband)\t" << tmp.signal_fit_status << " (Signal)" << endl;
  cout << "  Cov Status   : " << tmp.side_cov_status << " (Sideband)\t" << tmp.signal_cov_status << " (Signal)" << endl;
  cout << endl;
}

Result doFit(TH1 *input_sideband, TH1 *input_signal, bool saveme) {
  cout << "Performing New Fit" << endl;
  cout << "-------------------------------------------------------" << endl;
  cout << endl;

  Result res;

  double signal_region_count = input_signal->Integral(input_signal->FindBin(0), input_signal->FindBin(300));
  double sideband_region_count = input_sideband->Integral(input_sideband->FindBin(0), input_sideband->FindBin(300));

  RooRealVar x("x", "M(jj)", 0, 400, "GeV/c^{2}");
  RooRealVar norm("norm", "norm", 1.0, 0.0, 100.0);
  RooRealVar offset("offset", "offset", 10.849, 0, 30);
  RooRealVar power("power", "power", 2.4653, 0, 10);
  RooRealVar coeff1("coeff1", "coeff1", -0.0829876, -2, 2);
  RooRealVar coeff2("coeff2", "coeff2", 0.000133946, -1, 1);
//  RooThreshold thresh("thresh", "MN_FIT Threshold Function", x, RooArgList(norm, offset, power, coeff1, coeff2));

  RooRealVar gauss_mean("mean", "mean", 70, 50, 150);
  RooRealVar gauss_width("width", "width", 15, 0, 30);
  RooGaussian thresh("gauss", "Signal PDF", x, gauss_mean, gauss_width);

  RooDataHist data_sideband("data", "Yhea", x, input_sideband);
  RooDataHist data_signal("data", "Yhea", x, input_signal);

  // Set the regions
  x.setRange("fit", 0, 300);

  RooPlot *frame = 0x0;

  frame = x.frame();
  data_sideband.plotOn(frame, MarkerColor(1), LineColor(1));
  data_sideband.statOn(frame);
//  data_signal.plotOn(frame, MarkerColor(2), LineColor(2));
//  data_signal.statOn(frame);

  // Get the original integral
  RooAbsReal *preFitInt   = thresh.createIntegral(x);
  preFitInt->Print();

  // Generate the extended likelyhood stuff
  RooRealVar nbkg("nbkg", "Number of Background Events", 100, 0, 10000000);
  RooAddPdf bkgModel("BkgModel", "Background Model", RooArgList(thresh), RooArgList(nbkg));

  // Do the fit to the sidebands
  RooFitResult *fitResult_Sideband = bkgModel.fitTo(data_sideband, Extended(kTRUE), Range("fit"), Verbose(kTRUE), Save(kTRUE), Minos(kFALSE), PrintLevel(3));

  bkgModel.plotOn(frame, LineColor(kRed), LineStyle(2), Normalization(nbkg.getVal(), RooAbsReal::NumEvent));
  bkgModel.paramOn(frame, ShowConstants(kTRUE), Label("Fit Parameters"), Format("NEU", AutoPrecision(1)));

  int status = fitResult_Sideband->status();
  int covQual = fitResult_Sideband->covQual();
  cout << "The sideband fit status was: " << status << " with a covariance quality code of: " << covQual << endl;

  // Lock the shape parameters
  norm.setConstant(kTRUE);
  offset.setConstant(kTRUE);
  power.setConstant(kTRUE);
  coeff1.setConstant(kTRUE);
  coeff2.setConstant(kTRUE);

//  gauss_mean.setConstant(kTRUE);
//  gauss_width.setConstant(kTRUE);

  double pre_signal_norm = nbkg.getVal();
//  double pre_signal_norm = norm.getVal();

  // Do the fit to the signal region
  RooFitResult *fitResult_Signal = 0x0;
//  fitResult_Signal = bkgModel.fitTo(data_signal, Extended(kTRUE), Range("fit"), Verbose(kTRUE), Save(kTRUE), Minos(kFALSE), PrintLevel(3));

//  bkgModel.plotOn(frame, LineColor(kGreen));

  double post_signal_norm = nbkg.getVal();
//  double post_signal_norm = norm.getVal();

  if(fitResult_Signal != 0x0) {
    int status_signal = fitResult_Signal->status();
    int covQual_signal = fitResult_Signal->covQual();
    cout << "The signal fit status was: " << status_signal << " with a covariance quality code of: " << covQual_signal << endl;
  }

  norm.setConstant(kFALSE);
  offset.setConstant(kFALSE);
  power.setConstant(kFALSE);
  coeff1.setConstant(kFALSE);
  coeff2.setConstant(kFALSE);

  RooAbsReal *postIntSignal = thresh.createIntegral(x);
  postIntSignal->Print();

//  cout << "The post fit signal integral is             : " << postIntSignal->getVal() << endl;
  cout << "The post fit signal integral is             : " << post_signal_norm << endl;
  cout << "The number of events in the signal region is: " << signal_region_count << endl;
  cout << "The number of signal events is              : " << signal_region_count - postIntSignal->getVal() << endl;
  cout << "The pre norm was: " << pre_signal_norm << " and the post signal was: " << post_signal_norm << endl;

  TCanvas *can = new TCanvas("mycanvas");

  frame->SetTitle("M(jj) Fit with a Threshold Function");
  frame->Draw();

  if(saveme == true) {
    cerr << "Saving a plot!" << endl;
    can->SaveAs("test.eps");
  }

//  double frac_err = norm.getError() / norm.getVal();

  res.obs               = sideband_region_count;
  res.obs_err           = sqrt(res.obs);
  res.exp               = post_signal_norm;
  res.exp_err           = nbkg.getError();
//  res.exp_err           = res.exp * frac_err;
  res.delta             = res.obs - res.exp;
  res.pull              = res.delta / res.exp_err;
  res.side_fit_status   = fitResult_Sideband->status();
  res.side_cov_status   = fitResult_Sideband->covQual();
  res.signal_fit_status = 0;
  res.signal_cov_status = 3;
//  res.signal_fit_status = fitResult_Signal->status();
//  res.signal_cov_status = fitResult_Signal->covQual();
  res.norm              = norm.getVal();
  res.norm_err          = norm.getError();
  res.offset            = offset.getVal();
  res.offset_err        = offset.getError();
  res.power             = power.getVal();
  res.power_err         = power.getError();
  res.coeff1            = coeff1.getVal();
  res.coeff1_err        = coeff1.getError();
  res.coeff2            = coeff2.getVal();
  res.coeff2_err        = coeff2.getError();
  res.mean              = gauss_mean.getVal();
  res.mean_err          = gauss_mean.getError();
  res.width             = gauss_width.getVal();
  res.width_err         = gauss_width.getError();

  res.signal_histo      = (TH1*) input_signal->Clone();
  res.sideband_histo    = (TH1*) input_sideband->Clone();
  res.canvas            = (TCanvas*) can->Clone();

  return(res);
}
